#include "utils.hh"
#include "rational.hh"
#include <ctime>
#include "ellFixJ.hh"
#include "ell.hh"
#include "cuspidal.hh"
#include "stdio.h"
#include "stdlib.h"
#include <sstream>

mymap dbr,dbs,dbn,dbc,dbe;
map2mymap db2p;


void error(string msg) {
	cerr << msg << endl;
	exit(1);
}

Curve* init(string type){
	Curve* curve;
	if (type == "elliptic")
		return new Elliptic;
	if (type == "nodal") 
    return new Nodal;
	if (type == "cuspidal")
		return new Cuspidal;
	if (type == "fixJ")
		return new FixJ;
	if (type == "rational")
		return new Rational;
	cerr << "Invalid curve type" << endl;
	exit(1);
}


int main() {
	vector<int> cs;
	string type;
	int i;
	int dim,deg;
	char str[256];
	fgets(str,256, stdin);
	string command = str;
	istringstream iss(command);
	iss >> type;
	Curve* curve = init(type);

	iss >> dim;
	iss >> deg;

	if (dim > 5 && curve->GetAGenus() > 0)
		cerr << "Dimension higher than 5 for non-rational curves is not yet implemented";


	curve->SetData(dim,deg);
	iss >> i;
	cs.push_back(i);
	cs.push_back(0);
	while (iss >> i) {
		cs.push_back(i);
	}

	if (curve->IsSingular()) {
		if (cs.size() != dim + 2) 
			error("Invalid condition");
	}
	else
		if (cs.size() != dim + 1) 
			error("Invalid condition");
		

	if (curve->DimFamily(cs) != 0) 
		error("Incompatible constraint");

	cout << curve->Enumerate(cs) << endl;	
	delete(curve);
	return 0;	
	
}

