#include "curves.hh"
#include "utils.hh"


mpz_class ContractedComponent(Curve* c, int nodeCodim, vector<int> cs, vector<int> ao1, vector<int> ao2) {
	int dim = c->GetDim();
	mpz_class result(0);
  if (numSP(ao1) != 2) 
    return 0;
  int newCodim = nodeCodim + sum(ao1);
  if (newCodim > dim)
    return 0;

  cs[newCodim] +=1;
  for (unsigned int i=0; i < ao2.size();i++)
    cs[i] += ao2[i];
  
  return c->Enumerate(cs);
}

mpz_class ContractedComponentFirst(Curve* c2, Curve* c3, int nodeCodim1, int nodeCodim2, vector<int> cs, vector<int> ao1, vector<int> ao2, vector<int> ao3) {
	int dim = c2->GetDim();
	mpz_class result(0);
  if (numSP(ao1) != 2) 
    return 0;
  int newCodim1 = nodeCodim1 + sum(ao1);
  if (newCodim1 > dim)
    return 0;

  ao2[newCodim1] +=1;
  //for (unsigned int i=0; i < ao2.size();i++)
   // cs[i] += ao2[i];
	/*if (c3->GetGenus() == 1) {
		cout << "good" << endl;
		printVector(cs);
		printVector(ao2);
		printVector(ao3);
		cout << c3->GetDeg() << " " << c2->GetDeg() << " " << nodeCodim2 << endl;
  }*/
  return CharPair(c2,c3,nodeCodim2,cs,ao2,ao3);// c->Enumerate(cs);
}

mpz_class ContractedComponentMiddle(Curve* c1, Curve* c3, int nodeCodim1, int nodeCodim2, vector<int> cs, vector<int> ao1, vector<int> ao2, vector<int> ao3) {
	int dim = c1->GetDim();
	mpz_class result(0);
  if (numSP(ao2) != 1) 
    return 0;
  int newCodim = nodeCodim1 + nodeCodim2 + sum(ao2);
  if (newCodim > dim)
    return 0;

  //ao2[newCodim1] +=1;
  //for (unsigned int i=0; i < ao2.size();i++)
   // cs[i] += ao2[i];
  
  return CharPair(c1,c3,newCodim,cs,ao1,ao3);// c->Enumerate(cs);
}




mpz_class CharPair(Curve* c1, Curve* c2, int nodeCodim, vector<int> cs, vector<int> ao1, vector<int> ao2){
	int dim = c1->GetDim();
	//if (dim != c2->GetDim()) {
	//	cout << "Pair of curves have different ambient dimensions" << endl;
	//	exit(1);
	//}
	if (nodeCodim > dim)
		return 0;
	
	mpz_class result(0);	

	if (c1->GetAGenus() == 0 && c1->GetDeg() == 0) 
		return ContractedComponent(c2,nodeCodim,cs,ao1,ao2);
	if (c2->GetAGenus() == 0 && c2->GetDeg() == 0) 
		return ContractedComponent(c1,nodeCodim,cs,ao2,ao1);
	

	vector<int> cs1(dim+1,0); cs1[0]=-1;
  vector<int> cs2(dim+1,0);
	if (cs.size() > dim + 1) {
		if (c1->IsSingular()) {
			assert(cs.size() >= dim + 1);
			cs1.push_back(cs[dim+1]);
		}
		if (c2->IsSingular()) {
			assert(cs.size() >= dim + 1);
			cs2.push_back(cs[dim+1]);
		}
	}

	while (true) {
		int id =0; int j;
		while (id <= dim && cs1[id] == cs[id])
			id++;
		if (id == dim+1)
			break;
		rep(j,0,id)
			cs1[j] =0;
		cs1[id]++;
		mpz_class mult =1;
		rep(j,0,dim+1) {
			cs2[j] = cs[j] - cs1[j];
			
			mult *= combinations(cs1[j], cs[j]);
		}	
		

    for (unsigned int i=0; i <  ao1.size(); i++)
      cs1[i]  += ao1[i];
    for (unsigned int i=0; i <  ao2.size(); i++)
      cs2[i]  += ao2[i];

    mpz_class increment(1);
    increment *= CharPairSplit(c1,c2,nodeCodim,cs1,cs2);
		//printVector(cs1);
		//printVector(cs2);
		//cout << result << endl;
		increment *= mult;
    result += increment;

		for (unsigned int i=0; i <  ao1.size(); i++)  // return contraints1 to current state.
      cs1[i]  -= ao1[i];
  }

  return result;
}

mpz_class CharPairSplit(Curve* c1, Curve* c2, int nodeCodim, vector<int> cs1, vector<int> cs2) {
	int dim = c1->GetDim();
	int dim1= c1->DimFamily(cs1) +1 - nodeCodim;
  int dim2= c2->DimFamily(cs2) +1 - nodeCodim;


  if (dim1 <0 || dim2 <0 || dim1 + dim2 + nodeCodim < dim) 
    return 0;
  if (dim1 ==0 && nodeCodim == 0)
    return 0;
  if (dim2 ==0 && nodeCodim == 0)
    return 0;

  
  if (dim1 + dim2 + nodeCodim > dim ) {
    cout << "imcomplete collection at CharPairSplit"  <<  endl;
    exit(1);
  }

//		if (c1->GetDeg() == 1 && c2->GetDeg() == 3 && c2->GetGenus() == 1){

//	printVector(cs1);
//	printVector(cs2);
//	cout << dim1 << " " << dim2 << endl;
//	}


  if (dim1+nodeCodim>0)
    cs1[dim1+nodeCodim] +=1;
  if (dim2+nodeCodim>0)
    cs2[dim2+nodeCodim] +=1;

  
//	cout << "here" << endl;

  mpz_class num1, num2;
  num1 = c1->Enumerate(cs1);
  num2 = c2->Enumerate(cs2);
	
	mpz_class result = num1*num2;
	if (c1->GetDeg() == 2 && c1->GetAGenus() > 0)
		result/= 2;
	if (c2->GetDeg() == 2 && c2->GetAGenus() > 0)
		result/= 2;
	//if (c2->GetDeg() == 3)
		//cout << result << " ";
	  
  return result;
}

mpz_class CharCurvePairSplit(Curve* c1, Curve* c2, Curve* c3, int nodeCodim1, int nodeCodim2, vector<int> cs1, vector<int> cs2, vector<int> ao2, vector<int> ao3){
	
	int dim = c1->GetDim();
	//vector<int> dummy(dim+1,0);
	//if (c3->IsSingular())
	//	dummy.push_back(0);
	int dim1= c1->DimFamily(cs1) +1 - nodeCodim1;
  int dim2= c2->DimFamily(cs2) + c3->DimFamily(ao3)- dim  - sum(ao2) + numSP(ao2)+ 3 - nodeCodim1;
	//printVector(cs1);
	//printVector(cs2);
	//cout << dim1 << " " << dim2 << endl;

	/*if (c1->GetDeg() == 3 && c1->GetGenus() == 1 && cs1[2] == 8) {
		printVector(cs1);
		printVector(cs2);
		printVector(ao2);
		printVector(ao3);
		cout << endl;
		cout << dim1 << endl;
		cout << dim2 << endl;
	}*/



  if (dim1 <0 || dim2 <0 || dim1 + dim2 + nodeCodim1 < dim) 
    return 0;
  if (dim1 ==0 && nodeCodim1 == 0)
    return 0;
  if (dim2 ==0 && nodeCodim1 == 0)
    return 0;



  
  if (dim1 + dim2 + nodeCodim1 > dim ) {
    cout << "imcomplete collection at CharCurvePairSplit"  <<  endl;
		printVector(cs1);
		printVector(cs2);
		cout << c1->GetDeg() << " " << c2->GetDeg() << " " << c3->GetDeg() << " ";
		cout << dim1 << " " << dim2 << " " << nodeCodim1 << " "<< c1->GetGenus() << " " << endl;
		printVector(ao2);
		printVector(ao3);
    exit(1);
  }
	//		if (c1->GetDeg() == 3 && c1->GetGenus() == 1) {
	//	printVector(cs1);
	//	printVector(cs2);
	//	printVector(ao2);
	//	printVector(ao3);
	//	cout << endl;
//	}

  if (dim1+nodeCodim1>0)
    cs1[dim1+nodeCodim1] +=1;
  if (dim2+nodeCodim1>0)
    ao2[dim2+nodeCodim1] +=1;

  
//	cout << "here" << endl;

  mpz_class num1, num2;
  num1 = c1->Enumerate(cs1);
  num2 = CharPair(c2,c3,nodeCodim2,cs2,ao2,ao3);
	
	mpz_class result = num1*num2;
	if (c1->GetDeg() == 2 && c1->GetAGenus() > 0)
		result/= 2;
	//if (c2->GetDeg() == 2 && c2->GetAGenus() > 0)
	//	result/= 2;
	//if (c2->GetDeg() == 3)
		//cout << result << " ";
    
  return result;


}


mpz_class CharTriple(Curve* c1, Curve* c2, Curve* c3, int nodeCodim1, int nodeCodim2, vector<int> cs, vector<int> ao1, vector<int> ao2, vector<int> ao3){
	int dim = c1->GetDim();
	//if (dim != c2->GetDim()) {
	//	cout << "Pair of curves have different ambient dimensions" << endl;
	//	exit(1);
	//}
	if (nodeCodim1 > dim)
		return 0;
		
	//cout << "here" << endl;

	mpz_class result(0);	

	if (c1->GetAGenus() == 0 && c1->GetDeg() == 0) 
		return ContractedComponentFirst(c2,c3,nodeCodim1,nodeCodim2,cs,ao1,ao2,ao3);
	if (c2->GetAGenus() == 0 && c2->GetDeg() == 0) 
		return ContractedComponentMiddle(c1,c3,nodeCodim1,nodeCodim2,cs,ao1,ao2,ao3);
	if (c3->GetAGenus() == 0 && c3->GetDeg() == 0) {
		//cout << "here" << endl;

		return ContractedComponentFirst(c2,c1,nodeCodim2,nodeCodim1,cs,ao3,ao2,ao1);	
	}

	vector<int> cs1(dim+1,0); cs1[0]=-1;
  vector<int> cs2(dim+1,0);
	if (cs.size() > dim + 1) {
		if (c1->IsSingular())
			cs1.push_back(cs[dim+1]);
		if (c2->IsSingular() || c3->IsSingular())
			cs2.push_back(cs[dim+1]);
	}

	while (true) {
		int id =0; int j;
		while (id <= dim && cs1[id] == cs[id])
			id++;
		if (id == dim+1)
			break;
		rep(j,0,id)
			cs1[j] =0;
		cs1[id]++;
		mpz_class mult =1;
		rep(j,0,dim+1) {
			cs2[j] = cs[j] - cs1[j];
			
			mult *= combinations(cs1[j], cs[j]);
		}	
		

    for (unsigned int i=0; i <  ao1.size(); i++)
      cs1[i]  += ao1[i];
    //for (unsigned int i=0; i <  ao2.size(); i++)
     // cs2[i]  += ao2[i];

    mpz_class increment(1);
    increment *= CharCurvePairSplit(c1,c2,c3,nodeCodim1,nodeCodim2,cs1,cs2,ao2,ao3);
		//printVector(cs1);
		//printVector(cs2);
		//cout << result << endl;
		increment *= mult;
    result += increment;

		for (unsigned int i=0; i <  ao1.size(); i++)  // return contraints1 to current state.
      cs1[i]  -= ao1[i];
  }

  return result;

}


vector<int> Curve::Rearrange(vector<int> cs) {
	vector<int> result;
	int dim = GetDim();
	if (IsSingular()) {
		result.resize(dim+2);
		result[dim+1] = cs[cs.size()-1];
		for(int i=0; i <cs.size()-1 && i <= dim; i++)
			result[i] = cs[i];
	}
	else {
		result.resize(dim+1);
		for(int i=0; i <cs.size() && i <= dim; i++)
			result[i] = cs[i];
	}
	return result;
	
}
