#include "cuspidal.hh"

using namespace std;
extern mymap dbr, dbs, dbn, dbc, dbe;
extern map2mymap db2p;

mpz_class charCusp(int dim, int deg, vector<int> csc){
  int numHP =  csc[1];
  csc[1] =0;

  
  vector<int> data;
  data = readVector(csc);
  data.push_back(deg);
 
  if (dbc.find(data) != dbc.end()) 
    return dbc[data] * power(deg,numHP);


  mpz_class count =0;

	

  vector<int> ao1(dim+1,0);
  vector<int> ao2(dim+1,0);
  vector<int> cstp = csc; 
  int codimOfNode = -cstp[dim+1];
  if (codimOfNode > dim) 
    return 0;
  cstp.resize(cstp.size()-1);

  int degen =0;
  for (int degen =0; degen <= cstp[0] && degen <= dim; degen++) {
    vector<int> cscs = cstp;
    cscs[0] = cstp[0] - degen;
    for (int d=1; d  <  deg;d++)
      count += charBoundaryTwoJoints(dim,d,deg-d,cscs,codimOfNode,degen,ao1, ao2)*combinations(degen,cstp[0])*power(2,degen)*d*(deg-d);
  }
  cstp.push_back(-codimOfNode-1);
  count += charNodal(dim, deg, cstp)*2 * deg *2;

  cstp[dim+1] +=1;
  
  
  degen =0;
  mpz_class inc =0;
  for (degen =0; degen <= cstp[0] && degen <= dim; degen++) {
    vector<int> cscs = cstp;
    vector<int> cs_cs = cstp;
    cscs[0] = cstp[0] - degen;
    cs_cs[0] = cstp[0] - degen;
    for (int d=1; d+2 <= deg;d++)  { 
			Rational r1; Nodal n2;
			r1.SetData(dim,d);
			n2.SetData(dim,deg-d);
			inc = -CharPair(&r1,&n2,degen,cscs,ao1,ao2)*power(2,degen)*combinations(degen,cstp[0])*d*d*2;
      count += inc;
    }
    
    if (degen >0) { 
      cs_cs[dim+1] = cstp[dim+1] - degen;
      count += -charCusp(dim,deg,cs_cs)*combinations(degen,cstp[0])*deg*deg;
    }
  }
  cstp[2] +=1;
  count += -charNodal(dim,deg,cstp)*2;
  cstp[2] -=1;	//cout << charBoundaryRational(2,1,1,cs,0,ao1,ao2) << endl;*/
  if (count % (deg*deg) != 0)
    cout << "something wrong in countCusp" << endl;
  count /= deg*deg;
	dbc[data] = count;
  return count * power(deg,numHP);

}


mpz_class countCusp1(int dim, int deg, vector<int> csn){
  mpz_class count =0;

  vector<int> ao1(dim+1,0);
  vector<int> ao2(dim+1,0);
  vector<int> cs = csn; 
  int codimOfNode = -cs[dim+1];
  if (codimOfNode > dim) 
    return 0;
  cs.resize(cs.size()-1);

  int i=2;
  while (cs[i] == 0)
    i++;
  int p1 = i;
  cs[i] += -1;
  while (cs[i] ==0 )
    i++;
  int p2 = i;
  cs[i] += -1;
  ao1[p1] ++;
  ao2[p2] ++;


  for (int d=1; d  <  deg;d++)
    count += charBoundaryTwoJoints(dim,d,deg-d,cs,codimOfNode,0,ao1, ao2);
 
  cs.push_back(-codimOfNode-p1);
  cs[p2]++;
  count += charNodal(dim, deg, cs)*2;
  cs[p2]--;
  cs[dim+1] = -codimOfNode-p2;
  cs[p1]++;
  count += charNodal(dim, deg, cs)*2;
  cs[p1]--;

  cs[dim+1] =-codimOfNode;
  

  vector<int> ado1(dim+1,0);
  vector<int> ado2(dim+1,0); 
  ao1[p1]++;
  ao1[p2]++;
  mpz_class inc =0;
  for (int d=1; d+2 <= deg;d++)  { 
		Rational r1; Nodal n2;
		r1.SetData(dim,d);
		n2.SetData(dim,deg-d);
		inc = -CharPair(&r1,&n2,0,cs,ado1,ado2)*2;
    count += inc;
  }
    
  if (p1 + p2 <=dim) {
    cs[p1+p2] +=1;
    count += -charNodal(dim,deg,cs)*2;
    cs[p1+p2] -=1;
  }
  return count;

}


mpz_class planeTacnodalJoint(int deg1, int deg2, vector<int> cs1, vector<int> cs2, int nodeCodim) {
	if (nodeCodim == 1) {
		cout << "this case is not resolved :: planeTacnodalJoint :: nodeCodim = 1" << endl;
		exit(1);
	}
	int mult1 = (2*deg1-2)*deg2 + deg2*(deg2-1);
	int mult2 = (deg2-1)*(deg2-2);
	mpz_class numCurve2 = charRational(2,deg2,cs2);

	mpz_class result(0);
	//cs1.push_back(0);
	cs1[2] +=1;
	mpz_class smooth = mult1*charRational(2,deg1,cs1)*numCurve2; // if curve 2 were smooth.
	result += smooth;
	mpz_class throughNodes = mult2*charRational(2,deg1,cs1)*numCurve2; // through curve 2 nodes. 
	result -= throughNodes;
	cs1[2] -=1;
	cs1.push_back(-1);
	mpz_class nodesInC2 = 2*charNodal(2,deg1,cs1)*deg2*numCurve2;	//curve 2 pass through nodes of curve 1
	result -= nodesInC2;
	cs1.resize(3);
	//cout << result << endl;
	return result;
}

mpz_class countTacnodalJoint(int dim, int deg1, int deg2, vector<int> csr1, vector<int>csr2, int nodeCodim){
	int numHP1 = csr1[1];
	int numHP2 = csr2[1];
	csr1[1] = 0;
	csr2[1] = 0;
	mpz_class factor = power(deg1,numHP1)*power(deg2,numHP2);

  int dim1 = dimModSpace(dim,numSP(csr1),deg1)-sum(csr1) + 2;
  int dim2 = dimModSpace(dim,numSP(csr2),deg2)-sum(csr2) + 2;

  if (dim1 <2 || dim2 <2 || dim1 < nodeCodim || dim2 < nodeCodim) 
    return 0;
  if ((dim1 + dim2)  != 2*dim + nodeCodim + 1 ) {
    cout <<  "Counting 2 curves joined at 2 points and the dimensions don't add up" << endl;
    printConstraints(csr1);
    printConstraints(csr2);
    cout << deg1 << endl;
    cout << deg2 << endl;
    cout << "nodeCodim " << nodeCodim << endl;
    cout << "dim1 " << dim1 << endl;
    cout << "dim2 " << dim2 << endl;
    exit(1);
  }
	if (dim == 2) {
		if (dim1 > dim2)
	 	 	return planeTacnodalJoint(deg1,deg2,csr1,csr2,nodeCodim)*factor;
		else 
			return planeTacnodalJoint(deg2,deg1,csr2,csr1,nodeCodim)*factor;
	}
	

  mymap class1 = classOfLocusOfTwoPoints(dim,deg1,csr1);
  mymap class2 = classOfLocusOfTwoPoints(dim,deg2,csr2);
  mymap class0;
  vector<int> mono(3,0);
  mono[first_hyperplane] = nodeCodim;
  mono[second_hyperplane] = 0;
  mono[exceptional_divisor] = 1;
  class0[mono] = 1;
 
  mymap result = intersectionProduct(class0,class1,dim);
  result = intersectionProduct(result,class2,dim);

  mpz_class a=0;
  if (result.size() != 0)
    a = (*(result.begin())).second;
	if (a!=0) {
		printVector(csr1);
		printVector(csr2);
		cout << "dim1 " <<  dim1 << " deg1 " << deg1 << " dim2 " << dim2 << " deg2 " << deg2 << endl;
		cout << a << endl;
		printChowCycle(class1);
		printChowCycle(class2);
	}  

  return a*factor;

}

mpz_class countBoundaryTacnodal(int dim,int deg1,int deg2, vector<int> csr,int nodeCodim,vector<int> ao1,vector<int> ao2){
	mpz_class result(0);
	assert(csr[0] == 0);
  vector<int> csr1(dim+1,0); csr1[0]=-1;
  vector<int> csr2(dim+1,0);

	while (true) {
		int id =0; int j;
		while (id <= dim && csr1[id] == csr[id])
			id++;
		if (id == dim+1)
			break;
		rep(j,0,id)
			csr1[j] =0;
		csr1[id]++;
		mpz_class mult =1;
		rep(j,0,dim+1) {
			csr2[j] = csr[j] - csr1[j];
			mult *= combinations(csr1[j], csr[j]);
		} 
    
    for (unsigned int i=0; i <  ao1.size(); i++)
      csr1[i]  += ao1[i];
    for (unsigned int i=0; i <  ao2.size(); i++)
      csr2[i]  += ao2[i];
    

    mpz_class increment(1);
    increment *= countTacnodalJoint(dim,deg1,deg2,csr1,csr2,nodeCodim);
    increment *= mult;


		for (unsigned int i=0; i <  ao1.size(); i++)
      csr1[i]  -= ao1[i];
    

    result += increment;
  }

  return result;

}

mpz_class countCusp2(int dim, int deg, vector<int> csc){

	if (deg <= 2) 
		return 0;	
	mpz_class count(0);
	

  int nmSP = numSP(csc);
  
  if  (dimModSpaceNodal(dim,nmSP,deg) - 1< sum(csc)) {
    return 0;
  }
  if (dimModSpaceNodal(dim,nmSP,deg) - 1> sum(csc) ) {
    cout << "incomplete collection in countNodal" << endl;
    printdata(csc,deg,dim,0);
    exit(1);
  }

  
  if (nmSP == 2)
    return 1; 
  
  int i=2;
  while (csc[i] == 0)
    i++;
  int pivot = i;
  csc[i] += -1;
  while (csc[i] ==0 )
    i++;
  int s = i;
  csc[i] += -1;
  while (csc[i] ==0 )
    i++;
  int t = i;
  csc[i] += -1;
  
 
  
  vector<int> addons1(dim+1,0);
  vector<int> addons2(dim+1,0);
  vector<int> cs = csc;
  cs.resize(cs.size()-1);
  int codimOfNode = -csc[dim+1];
  if (codimOfNode > dim)
    return 0;

  addons1[pivot-1] += 1;
  addons1[1] +=1;
  addons2[s] +=1;
  addons2[t] +=1;
  for (int d=1; d<= deg-1;d++){
		Cuspidal c1; c1.SetData(dim,d);
		Rational r2; r2.SetData(dim,deg-d);
		mpz_class inc = 0;
		inc += -CharPair(&c1,&r2,0,csc,addons1,addons2);
		c1.SetDeg(deg-d);
		r2.SetDeg(d);
		inc += -CharPair(&r2,&c1,0,csc,addons1,addons2);
		cout << inc << " ";
		count +=inc;
		    

    if (d!= deg) {
			
      //count += -charBoundaryTwoJoints(dim,d,deg-d,cs,0,codimOfNode,addons1, addons2);
			mpz_class inc = -countBoundaryTacnodal(dim,d,deg-d,cs,codimOfNode,addons1,addons2);
			cout << inc << " ";
			count += 2*inc;
		}
  }


  addons1[1] -=1;
  addons1[t] +=1;
  addons2[1] +=1;
  addons2[t] -=1;

   
  for (int d=0; d<=deg;d++){
		Cuspidal c1; c1.SetData(dim,d);
		Rational r2; r2.SetData(dim,deg-d);
		mpz_class inc = 0;
		inc += CharPair(&c1,&r2,0,csc,addons1,addons2);
		c1.SetDeg(deg-d);
		r2.SetDeg(d);
		inc += CharPair(&r2,&c1,0,csc,addons1,addons2);
		count +=inc;
		cout << inc << " cus&rat ";
    if (d!=0 && d!= deg) {
      //count += charBoundaryTwoJoints(dim,d,deg-d,cs,0,codimOfNode,addons1, addons2);
			mpz_class inc = countBoundaryTacnodal(dim,d,deg-d,cs,codimOfNode,addons1,addons2);
			cout << inc <<  " ";
			count += 2*inc;
		}
  }
 
	
  return count;

}

mpz_class Cuspidal::Enumerate(vector<int> cs){
	cs = Rearrange(cs);
	return charCusp(GetDim(),GetDeg(),cs);
}


int Cuspidal::DimFamily(vector<int> cs){
	int dim = GetDim();
	int deg = GetDeg();
	return dimModSpaceNodal(dim,numSP(cs),deg) - sum(cs) - 1;

}
