#include "ellFixJ.hh"
#include <fstream>

using namespace std;

extern mymap dbr, dbs, dbn, dbc, dbe;
extern map2mymap db2p;

mymap classOfLocusOfTwoPoints(int dimSpace,int deg, vector<int> csr){

	vector<int> data;
  data = readVector(csr);
  data.push_back(deg);
	if (db2p.find(data) != db2p.end())
		return db2p[data];


  int dimClass = dimModSpace(dimSpace,numSP(csr),deg)- sum(csr) + 2;
  int codimClass = 2*dimSpace - dimClass;
  mymap result;


  // no e part:
  int minDimK = max(0, codimClass - dimSpace) ;
  int maxDimK = min(dimSpace, codimClass);


  for  (int dimK= minDimK ; dimK <= maxDimK; dimK++) {
    int dimH = codimClass - dimK;
    vector<int> mono(3,0);
    vector<int> cstp = csr;
    mono[first_hyperplane] = dimH;
    mono[second_hyperplane] = dimK;
    mono[exceptional_divisor] = 0;
    cstp[dimSpace-dimH]++;
    cstp[dimSpace-dimK]++;
    result[mono] = charRational(dimSpace,deg,cstp);
    if (result[mono] == 0)
      result.erase(mono);
  }

  // with e part:
  int minDimE = max(1, codimClass - dimSpace);
  int maxDimE = min(dimSpace-1,codimClass);
  vector<mpz_class> intermediateResults (dimSpace,0);
  vector<mpz_class> previousResults(dimSpace,0);
  for ( int dimE = maxDimE; dimE >=  minDimE;dimE--) {
    vector<int> cstp = csr;
    int dimH = codimClass - dimE;
    intermediateResults[dimSpace-dimE] = specialChar(dimSpace, deg, cstp, dimSpace-dimH, dimSpace-dimE-1);
    vector<int> mono (3,0);
    mono[first_hyperplane] = dimH;
    mono[second_hyperplane] = 0;
    mono[exceptional_divisor] = dimE;
    result[mono] =0;
    for (int j=0; j <= dimSpace-dimE-1;j++) 
      result[mono] += intermediateResults[dimSpace-dimE-j]*combinations(j,dimSpace-dimE-1)* power(2,j)*sign(dimSpace-dimE-1-j);
    for (int j=1; j <= maxDimE-dimE;j++)
      result[mono]+= -previousResults[dimE+j]*topMonomial(dimSpace+j,dimSpace);
    result[mono] *= sign(dimSpace-1);
    previousResults[dimE] = result[mono];
    if (result[mono] == 0)
      result.erase(mono);
  }
	db2p[data] = result;
  return result;

}
    


mpz_class charTwoJoints(int dim,int deg1,int deg2, vector<int> csr1, vector<int> csr2, int nodeCodim1, int nodeCodim2){
	int numHP1 = csr1[1];
	int numHP2 = csr2[1];
	csr1[1] = 0;
	csr2[1] = 0;


  int dim1 = dimModSpace(dim,numSP(csr1),deg1)-sum(csr1) + 2;
  int dim2 = dimModSpace(dim,numSP(csr2),deg2)-sum(csr2) + 2;

  if (dim1 <2 || dim2 <2 || dim1 < nodeCodim1 + nodeCodim2 || dim2 < nodeCodim1+nodeCodim2) 
    return 0;
  if ((dim1 + dim2)  != 2*dim + nodeCodim1 + nodeCodim2) {
    cout <<  "Counting 2 curves joined at 2 points and the dimensions don't add up" << endl;
    printConstraints(csr1);
    printConstraints(csr2);
    cout << deg1 << endl;
    cout << deg2 << endl;
    cout << "nodeCodim1 " << nodeCodim1 << endl;
    cout << "nodeCodim2 " << nodeCodim2 << endl;
    cout << "dim1 " << dim1 << endl;
    cout << "dim2 " << dim2 << endl;
    exit(1);
  }
 
  mymap class1 = classOfLocusOfTwoPoints(dim,deg1,csr1);
  mymap class2 = classOfLocusOfTwoPoints(dim,deg2,csr2);
  mymap class0;
  vector<int> mono(3,0);
  mono[first_hyperplane] = nodeCodim1;
  mono[second_hyperplane] = nodeCodim2;
  mono[exceptional_divisor] = 0;
  class0[mono] = 1;
 
  mymap result = intersectionProduct(class0,class1,dim);
  result = intersectionProduct(result,class2,dim);

  mpz_class a=0;
  if (result.size() != 0)
    a = (*(result.begin())).second;
  
  return a*power(deg1,numHP1)*power(deg2,numHP2);

}

mpz_class charBoundaryTwoJoints(int dim, int deg1,int deg2, vector<int> csr, int nodeCodim1, int nodeCodim2, vector<int> addons1, vector<int> addons2){

 
  mpz_class result(0);

  vector<int> csr1(dim+1,0); csr1[0]=-1;
  vector<int> csr2(dim+1,0);

	while (true) {
		int id =0; int j;
		while (id <= dim && csr1[id] == csr[id])
			id++;
		if (id == dim+1)
			break;
		rep(j,0,id)
			csr1[j] =0;
		csr1[id]++;
		mpz_class mult =1;
		rep(j,0,dim+1) {
			csr2[j] = csr[j] - csr1[j];
			mult *= combinations(csr1[j], csr[j]);
		} 
    
    for (unsigned int i=0; i <  addons1.size(); i++)
      csr1[i]  += addons1[i];
    for (unsigned int i=0; i <  addons2.size(); i++)
      csr2[i]  += addons2[i];
    

    mpz_class increment(1);
    increment *= charTwoJoints(dim,deg1,deg2,csr1,csr2,nodeCodim1,nodeCodim2);
    increment *= mult;

		for (unsigned int i=0; i <  addons1.size(); i++)
      csr1[i]  -= addons1[i];
    

    result += increment;
  }

  return result;
}


mpz_class countNodal(int dim, int deg, vector<int> csn){
  mpz_class count(0);


  int nmSP = numSP(csn);
  
  if  (dimModSpaceNodal(dim,nmSP,deg) < sum(csn)) {
    return 0;
  }
  if (dimModSpaceNodal(dim,nmSP,deg) > sum(csn) ) {
    cout << "incomplete collection in countNodal" << endl;
    printdata(csn,deg,dim,0);
    exit(1);
  }

  
  if (nmSP == 2)
    return 1; 
  
  int i=2;
  while (csn[i] == 0)
    i++;
  int pivot = i;
  csn[i] += -1;
  while (csn[i] ==0 )
    i++;
  int s = i;
  csn[i] += -1;
  while (csn[i] ==0 )
    i++;
  int t = i;
  csn[i] += -1;
  
 
  
  vector<int> addons1(dim+1,0);
  vector<int> addons2(dim+1,0);
  vector<int> cs = csn;
  cs.resize(cs.size()-1);
  int codimOfNode = -csn[dim+1];
  if (codimOfNode > dim)
    return 0;

  addons1[pivot-1] += 1;
  addons1[1] +=1;
  addons2[s] +=1;
  addons2[t] +=1;
  for (int d=1; d<= deg;d++){
		Nodal n1; n1.SetData(dim,d);
		Rational r2; r2.SetData(dim,deg-d);
		mpz_class inc = 0;
		inc += -CharPair(&n1,&r2,0,csn,addons1,addons2);
		n1.SetDeg(deg-d);
		r2.SetDeg(d);
		inc += -CharPair(&r2,&n1,0,csn,addons1,addons2);
		count +=inc;
		    

    if (d!= deg)
      count += -charBoundaryTwoJoints(dim,d,deg-d,cs,0,codimOfNode,addons1, addons2);
  }


  addons1[1] -=1;
  addons1[t] +=1;
  addons2[1] +=1;
  addons2[t] -=1;

   
  for (int d=0; d<=deg;d++){
		Nodal n1; n1.SetData(dim,d);
		Rational r2; r2.SetData(dim,deg-d);
		mpz_class inc = 0;
		inc += CharPair(&n1,&r2,0,csn,addons1,addons2);
		n1.SetDeg(deg-d);
		r2.SetDeg(d);
		inc += CharPair(&r2,&n1,0,csn,addons1,addons2);
		count +=inc;
    if (d!=0 && d!= deg)
      count += charBoundaryTwoJoints(dim,d,deg-d,cs,0,codimOfNode,addons1, addons2);
  }
 
	
  return count;

}


mpz_class charNodal(int dim, int deg, vector<int> csn){

  if (deg <2)
    return 0;

  if (deg == 2)
    return charDoubleCover(dim, csn);


  mpz_class result(0);

 
  int numHP =  csn[1];
  csn[1] =0;

  
  vector<int> data;
  data = readVector(csn);
  data.push_back(deg);
 
  if (dbn.find(data) != dbn.end()) 
    return dbn[data] * power(deg,numHP);

  if (csn[0] == 0) {
    dbn[data] = countNodal(dim, deg, csn);
    return dbn[data]*power(deg,numHP);
  }

 
  
  csn[0] -= 1;
  csn[2] += 1;
  result += charNodal(dim,deg,csn)*(deg-1)*2;
  csn[2] -= 1;

  int tangency= csn[0];


  int codimOfNode = -csn[dim+1];
  if (codimOfNode > dim)
    return 0;
  vector<int> cs = csn;
  cs.resize(cs.size()-1);

  for (int d=1; 2*d <=deg; d++) {
    csn[0]= tangency;
    int degen = 0;
    while (degen <= tangency && degen <= dim) {
      vector<int> addons1;
      vector<int> addons2;
			Rational r1,r2; Nodal n1,n2;
			r1.SetData(dim,d); r2.SetData(dim,deg-d);
			n1.SetData(dim,d); n2.SetData(dim,deg-d);
			mpz_class increment = CharPair(&r1,&n2,degen,csn,addons1,addons2) + CharPair(&n1,&r2,degen,csn,addons1,addons2);      
			//mpz_class increment = charBoundaryNodal(dim,d,deg-d,csn,degen,addons1,addons2);
      increment *= combinations(degen,tangency)*power(2,degen+1)*d*(deg-d);
      if (d == deg -d)
         increment /= 2;
      result += increment;
      
      degen++;
      csn[0]--;
    }
    
  }
  csn[0] = tangency;
  




  for (int d=1; 2*d <= deg;d++){
    cs[0] = tangency;
    int degen =0; // number of planes passing through a node.
    while ( degen <= tangency  &&  degen<= dim) {
 
      
      vector<int> addons1;
      vector<int> addons2;
      mpz_class increment = charBoundaryTwoJoints(dim,d,deg-d,cs,degen,codimOfNode,addons1, addons2);
    
      increment *=combinations(degen,tangency)*power(2,degen+1)*d*(deg-d);
    
      if (d ==deg-d)
	increment /=2;
      result += increment;

      degen++;
      cs[0]--;
    }
  }
  
  if (result % (deg*2) != 0) {
    cout << "deg is " << deg << endl;
    cout << "tangency is " << tangency +1 <<endl;
    printVector(csn);
  }
  result /= deg*2;
	dbn[data] = result;
	result *= power(deg,numHP);

  
	//if (dim == 2) 
	//	result /= combinations(2,deg -1);
  

  return result;
  

}

mpz_class charDoubleCover(int dim, vector<int> csn){
 
  int codimOfNode = -csn[dim+1];
  int tangency = csn[0];
  if (codimOfNode < 1 ||  codimOfNode > dim || tangency < 2)
    return 0;
  mpz_class result(0);
  vector<int> cs(dim+1,0);
  for(unsigned int i=1; i < cs.size();i++)
    cs[i] = csn[i];
  int f  = numSP(cs);
  cs[codimOfNode]++;
   //partion the set of tangency conditions into two subsets
  for (int num1 =1; num1 < tangency; num1 ++) {
    int num2 = tangency - num1;
    if (num1 <= dim && num2 <= dim) {
      cs[num1]++;
      cs[num2]++;
			mpz_class mult = combinations(num1,tangency)*power(2,f);
      result += charRational(dim,1,cs)*mult;
      cs[num1]--;
      cs[num2]--;
    }
  }
  result /=2;


  return result;
}

mpz_class charFixJ(int dim, int deg, vector<int> csfj){
  mpz_class result =0;
  vector<int> cs = csfj; 
  int tangency = csfj[0];
  cs.push_back(0);
  for (int degen =0; degen <= tangency && degen <= dim; degen++) {
    cs[0] = tangency - degen;
    cs[dim+1]= -degen;
    result += charNodal(dim,deg,cs)* power(2,degen) * combinations(degen,tangency);
  }
  return result;
}



/*-------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
--------------------------------------------Test Functions-----------------------------------------------------
--------------------------------------------Test Functions-----------------------------------------------------
--------------------------------------------Test Functions-----------------------------------------------------
--------------------------------------------Test Functions-----------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------
-------------------------------------------------------------------------------------------------------------*/


void testTopMonomial(){
  int dimSpace = 5;
  vector<int> mono(3,0);
  mono[first_hyperplane] = 0;
  mono[second_hyperplane] = 0;
  mono[exceptional_divisor] = 2*dimSpace - mono[first_hyperplane];
 
}


void testIntersectionProduct(){
  int dimSpace = 3;
  string expression1 = "";
  string expression2 = "";
  mymap cycle1 = readCycleFromString(expression1);
  mymap cycle2 = readCycleFromString(expression2);
  printChowCycle(cycle1);
  printChowCycle(cycle2);
  mymap productCycle = intersectionProduct(cycle1,cycle2,dimSpace);

  printChowCycle(productCycle);
  
    
  
}


  
/*----------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------Functions that involve the special class--------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
------------------------------------------------------------------------------
----------------------------------------------------------------------------*/


mpz_class specialChar(int dim, int deg, vector<int> css, int markCodim, int numClass){
  if (numClass >3){
    cout <<"too many special classes in specialChar" << endl;
    exit(1);
  }

  if (markCodim >= dim+1)
     return 0;

  if (numClass == 0) {
    if (markCodim != 0) 
      css[markCodim] ++;
    return charRational(dim,deg,css);
  }
	int numHP = css[1];
  css[1] = 0;


	vector<int> data;
	data = readVector(css);
	data.push_back(deg);
	data.push_back(markCodim);
	data.push_back(numClass);
	if (dbs.find(data) != dbs.end()) 
    return dbs[data] * power(deg,numHP);

  

  mpz_class result = 0;
  css[2] +=1;
  result += specialChar(dim,deg,css,markCodim,numClass-1);
  css[2] -=1;
    
  result += specialChar(dim,deg,css,markCodim+1,numClass-1)*2*deg*(deg-1);
        
  
  
  int tangency= css[0];


  for (int d=1; d<deg;d++){
    css[0] = tangency;
    int degen =0; // number of planes passing through a node.
    int multipleOfE2 = 0;
    while (css[0] >=0  && degen <= dim) {
      
      vector<int> addons1;
      vector<int> addons2;

      mpz_class inc = specialCharBoundary(dim,deg-d,d,css,degen,addons1,addons2,markCodim,numClass-1);
      inc *=combinations(degen,tangency)*power(2,degen)*(deg-d)*(deg-d);
      result += inc;
      degen++;
      css[0]--;
    }
    css[0] = tangency;
    if (d <=deg-d) {
      multipleOfE2 += d*d+(deg-d)*(deg-d) - deg*deg;
      mpz_class inc= intersectionWithE2(dim,deg-d,d,css,markCodim,numClass-1)*multipleOfE2;
      if(d == deg-d)
				inc /= 2;
      result += inc;
    }
  }

	
  if (result % (deg*deg) != 0) {
    printVector(css);
    cout <<  "markCodim " << markCodim << endl;
    cout << "numClass " << numClass << endl;
  }
  result /= deg*deg;
	dbs[data] = result;

	result *= power(deg,numHP);
  
  return result;
 

}


mpz_class specialCharOneJoint(int dim, int deg1, int deg2, vector<int> csr, vector<int> css, int nodeCodim, int markCodim, int numClass){
  

  int dim1= -sum(csr) + dimModSpace(dim,numSP(csr),deg1)+1;
  int dim2= -sum(css) + dimModSpace(dim,numSP(css),deg2)+1 - markCodim + (markCodim!=0 || numClass!=0) - numClass;


  if (dim1 - nodeCodim  <0 || dim2 - nodeCodim < 0 || dim1 < 1 || dim2 < 1) 
    return 0;

  
  if (dim- dim1 + dim- dim2 + nodeCodim < dim ) {
    cout << "imcomplete collection at charOneJoint"  <<  endl;
    exit(1);
  }
  
  if (dim-dim1 + dim-dim2 + nodeCodim > dim) 
    return 0;


  if(dim1 >0)
    csr[dim1]++;
  if(dim2 >0)
    css[dim2]++;
  
  mpz_class num1, num2;
  num1 = charRational(dim,deg1,csr);
 
  num2 = specialChar(dim,deg2,css,markCodim,numClass);
      
  return num1*num2;



}



mpz_class specialCharBoundary(int dim, int deg1, int deg2,vector<int> cs, int nodeCodim, vector<int> addons1, vector<int> addons2,int markCodim, int numClass){
  
  
  
  mpz_class result(0);
 
  vector<int> csr(dim+1,0); csr[0]=-1;
  vector<int> css(dim+1,0);


	while (true) {
		int id =0; int j;
		while (id <= dim && csr[id] == cs[id])
			id++;
		if (id == dim+1)
			break;
		rep(j,0,id)
			csr[j] =0;
		csr[id]++;
		mpz_class mult =1;
		rep(j,0,dim+1) {
			css[j] = cs[j] - csr[j];
			
			mult *= combinations(csr[j], cs[j]);
		}	

   
    for (unsigned int i=0; i <  addons1.size(); i++)
      csr[i]  += addons1[i];
    for (unsigned int i=0; i <  addons2.size(); i++)
      css[i]  += addons2[i];

    mpz_class increment(1);
    increment *= specialCharOneJoint(dim,deg1,deg2,csr,css,nodeCodim,markCodim,numClass);
    increment *= mult;
  
		for (unsigned int i=0; i <  addons1.size(); i++)
      csr[i]  -= addons1[i];
 
    result += increment;
  }

  return result;
}


mpz_class intersectionWithXa(int ,int ,int ,vector<int> ,vector<int> ,int ,int);

mpz_class intersectionWithE2(int dim, int deg1,int deg2, vector<int> cstt, int markCodim, int numClass){
  mpz_class result =0;
  if (numClass == 0) 
    return 0;
  if (numClass == 1) {
    int tangency = cstt[0]; 
    vector<int> cs = cstt;
    mpz_class increment = 0;
    vector<int> addons1;
    vector<int> addons2;
    for (int degen =0; degen <= tangency; degen++) {
      cs[0] = tangency - degen;
			Rational r1,r2;
			r1.SetData(dim,deg1);	
			r2.SetData(dim,deg2);
      increment = CharPair(&r1,&r2,markCodim+degen,cs,addons1,addons2) * power(2,degen) * combinations(degen,tangency);
    result +=  increment;
    }
    return result;
  }
  if (numClass == 2) {
   
    vector<int> cs1(dim+1,0);
    vector<int> cs2(dim+1,0);
    
    int tangency = cstt[0];
    vector<int> cs = cstt;
    for (int degen = 0; degen <= tangency; degen++) {
       cs[0] = tangency - degen;
       int numCase = numberOfCases(cs);
       for (int caseNum = 0; caseNum < numCase; caseNum++) {
         int partition = caseNum;
         int level = numCase;
         mpz_class mult = combinations(degen,tangency)*power(2,degen);
         for (int codim =0; codim <= dim; codim++) {
	   level = level / (cs[codim] + 1);
	   div_t result = div(partition, level);
	   cs1[codim] = result.quot;
	   cs2[codim] = cs[codim]-cs1[codim];
	   partition = result.rem;
	   mult *= combinations(cs1[codim], cs[codim]);
         }
         mpz_class increment(0);
         increment += intersectionWithXa(dim,deg1,deg2,cs1,cs2,markCodim+degen,numClass);
         increment += intersectionWithXa(dim,deg2,deg1,cs2,cs1,markCodim+degen,numClass);
         increment *= mult;
         result += increment;
      }    
    }

 
  return result;
  }


  cout << "something wrong in intersectionWithE2" << endl;
  return 0;

}

mpz_class intersectionWithXa(int dim,int deg1,int deg2,vector<int> css,vector<int> csr,int markCodim,int numClass){
  if (markCodim > dim) return 0;
  int dim1= -sum(css) + dimModSpace(dim,numSP(css),deg1)+1;
  int dim2= -sum(csr) + dimModSpace(dim,numSP(csr),deg2)+1;

  if (dim1 - markCodim  <1 || dim2 - markCodim < 0 || dim1 < 1 || dim2 < 1) 
    return 0;

  
  if (dim- dim1 + dim- dim2 + markCodim < dim-1 ) {
    cout << "imcomplete collection at intersectionWithXa"  <<  endl;
    exit(1);
  }

  if (dim - dim1 + dim - dim2 + markCodim > dim-1) {
    cout << "hello" << endl;
    return 0;
  }

  // dim1 + dim2 = dim + markCodim +1
  // dim1 >= markCodim


  if (dim2 == dim + 1) // hence dim1 = markCodim
    return 0;

  
  mpz_class c1 = specialChar(dim,deg1,css,dim1-1,1);
  if (dim2!=0)
    csr[dim2]++;
  mpz_class c2 = charRational(dim,deg2,csr);
  return c1*c2;

}

int Nodal::DimFamily(vector<int> cs){
	int dim = GetDim();
	int deg = GetDeg();
	return dimModSpaceNodal(dim,numSP(cs),deg) - sum(cs);

}

mpz_class Nodal::Enumerate(vector<int> cs){
	cs = Rearrange(cs);
	return charNodal(GetDim(), GetDeg(), cs);
}

mpz_class FixJ::Enumerate(vector<int> cs){
	cs = Rearrange(cs);
	return charFixJ(GetDim(), GetDeg(), cs);
}

int FixJ::DimFamily(vector<int> cs){
	int dim = GetDim();
	int deg = GetDeg();
	return dimModSpaceNodal(dim,numSP(cs),deg) - sum(cs);

}
