#include "utils.hh"

void combinations(mpz_t result, int k, int n) {
	int i;	
	assert(k <= n);
	if (2*k > n)
		k = n-k;
	rep(i,0,k) {
		mpz_mul_ui(result,result,(uli)(n-i));
		mpz_divexact_ui(result,result,(uli)(i+1));		
	}	
}

int **split(int* vector, int length){
	return 0;	
}

mpz_class combinations(int k, int n)  {
	mpz_class res = 1;
	int i;
	assert(k<=n);
	if (2*k > n)
		k = n-k;
	rep(i,0,k) {
		res*= n-i;
		res/= i+1;
	}
	return res;
		
}


mymap readCycleFromString(string expression){
  mymap cycle;
  
  if (expression[0] != '-')
    expression.insert(0,"+");
  expression.insert(1," ");
  unsigned int pos;
  cout << "expression:" <<expression << endl;
  
  while (true) {
    pos = expression.find_first_of("+-",2);
    if( pos == expression.npos )
      pos = expression.size() -1;
    else 
      pos = pos -2;
    string mono = expression.substr(2,pos-1);
    unsigned int var = mono.find_first_of("hke");
    string coef = mono.substr(0,var);
		mpz_class coefficient(coef.size() == 0 ? "1" : coef,10);
    if (expression[0] == '-')
      coefficient = -coefficient;
    vector<int> varPositions;
    varPositions.push_back(var);
   

    while(true) {
      var = mono.find_first_of("hke",var+1);
      if (var == mono.npos)
	break;
      varPositions.push_back(var);
    }
    vector<int> monomial(3,0);
    
    for (unsigned int i=0; i < varPositions.size(); i++){
      string pow;
          
      if ( i != varPositions.size() -1 )
	pow = mono.substr(varPositions[i]+2,varPositions[i+1]-varPositions[i]-2 );
      else
	pow = mono.substr(varPositions[i]+2,mono.size() - varPositions[i]-2);
      int power = atoi(pow.c_str());

      switch (mono[varPositions[i]]){
      case 'h': 
	monomial[first_hyperplane] = power;
	break;
      case 'k':
	monomial[second_hyperplane] = power;
	break;
      case 'e':
	monomial[exceptional_divisor] = power;
	break;
      }
    }
    monomial = reducedForm(monomial);
    cycle[monomial] = coefficient;
    if (pos+2 > expression.size())
      break;
    expression = expression.substr(pos+2);
    }

  return cycle;
}



void printChowCycle(mymap chowCycle){
  if (chowCycle.size() == 0) {
    cout << "This cycle is empty" << endl;
    return;
  }
  mymap::iterator it;
  for(it = chowCycle.begin(); it!=chowCycle.end();it++){
    mpz_class coefficient = (*it).second;
    if (it != chowCycle.begin())
      cout << " ";
  
    if (it!= chowCycle.begin() && coefficient > 0)
      cout <<"+";

    if (coefficient < 0){
      cout << "-";
      coefficient = -coefficient;
    }
    
    if (it!= chowCycle.begin())
      cout <<" ";
    if (coefficient !=1)
      cout << coefficient;
    vector<int> mono = (*it).first;
    if (mono[first_hyperplane]  != 0)
      cout <<"h^" << mono[first_hyperplane];
    if (mono[second_hyperplane] != 0)
      cout <<"k^" << mono[second_hyperplane];
    if (mono[exceptional_divisor] != 0)
      cout <<"e^" << mono[exceptional_divisor];
  } 
  cout << endl;
}

int factorial (int n){ 
  if (n<0) {
    cout << "Negative factorial" << endl;
    exit(1);
  }

  ;  if (n ==0 ) return 1;
  return n* factorial(n-1);
}

int dimModSpace(int r, int numMark, int deg) {
  if (r < 2 || deg <0) {
    cout <<  numMark;
    cout << "Absurd hypothesis" <<endl;
    exit(1); 
  }
  if (deg == 0) return r;
  if (deg == 1)
    return numMark + 2*r -2;
  return dimModSpace(r, numMark, deg-1) + 2*r-2 + 3-r;
}

int dimModSpaceNodal(int r, int numMark, int deg) {
  return dimModSpace(r,numMark,deg) - (r-2);

}

int dimModSpaceEll(int r, int numMark, int deg) {
  return dimModSpace(r,numMark,deg) - (r-3);
}

void printConstraints(vector<int> constraints){
  for (unsigned int i = 0; i < constraints.size(); i++)
    cout << constraints[i] <<  " ";
  cout << endl;

}

void printVector(vector<int> vects){
  for(unsigned int i=0; i< vects.size();i++)
    cout << vects[i] << " ";
  cout << endl;
}

void printdata(vector<int> cs, int deg, int dim, int num){
  printConstraints(cs);
  cout << "degree " << deg << " dimension " << dim << " number " << num <<endl;
}


int numberOfCases(vector<int> constraints){
  int numCase = 1;
  for (unsigned int i=0;  i < constraints.size(); i++)
    numCase *= constraints[i] + 1;
  return numCase;

}


vector<int> readVector(vector<int> nums){
  vector<int> result;
  for (unsigned int i=0; i < nums.size(); i++)
    if (i !=1)
      result.push_back(nums[i]);
  return result;
}


int codim(vector<int> cs) {
	int result =0;
	for(int i=1; i < (int)cs.size();i++)
		result += cs[i]*(i-1);
	result += cs[0];
	return result;
}

int sum(vector<int> constraints) { // the total of codimensions in a collection of subspaces.
  int s = 0;
  for (unsigned int i=1; i  < constraints.size(); i++)
    if (constraints[i] >= 0)
      s += i*constraints[i];
    else
      s += -constraints[i];
  s += constraints[0];
  return s;
}

int numSP(vector<int> constraints) {// number of incident subspaces in the collection
  int s = 0;
  for (unsigned int i=1; i< constraints.size();i++)
    if (constraints[i] > 0)    
      s+= constraints[i];
  return s;
}

int sumVector(vector<int> vects) {
  int s =0;
  for (unsigned int i=0; i<vects.size();i++)
    s+= vects[i];
  return s;
}


int power(int a, int n ) {
  if (n<0)  {
    cout << "negative power" << endl;
    exit(1);
  }
  if (n==0) return 1;
  if (n==1) return a;
  int  m  = n/2;
  return power(a,m)*power(a,n-m);
  
}






void cleanUp(mymap& chowClass, int dimSpace){
  mymap::iterator it;
  bool needCleanUp;
  int loopCount =0;
  while (true) {
    loopCount ++;
    needCleanUp = false;
    for (it = chowClass.begin(); it != chowClass.end(); it++)  {
      vector<int> monomial = (*it).first;
      mpz_class coefficient = (*it).second;
      int dimClass =  sumVector(monomial);
      if (monomial[exceptional_divisor] >= dimSpace && dimClass <= 2*dimSpace && coefficient !=0) {
	chowClass.erase(it);
	for (int j=1; j< dimSpace;j++){
	  vector<int> mono (3,0);
	  mono[first_hyperplane] = monomial[first_hyperplane] +j;
	mono[exceptional_divisor] = monomial[exceptional_divisor] - j;
	mono = reducedForm(mono);
	if (chowClass.find(mono) == chowClass.end())
	  chowClass[mono] = 0;
	chowClass[mono] += coefficient*combinations(j,dimSpace+1)*sign(j-1);
	}
	for (int k=0; k <=dimSpace;k++) {
	  vector<int> mono(3,0);
	  mono[first_hyperplane] = monomial[first_hyperplane] + k;
	  mono[second_hyperplane] = dimSpace-k;
	  mono[exceptional_divisor] = monomial[exceptional_divisor] -dimSpace;
	  mono = reducedForm(mono);
	  if (chowClass.find(mono)== chowClass.end())
	    chowClass[mono]=0;
	  chowClass[mono] += coefficient*sign(dimSpace-1);
	}
	needCleanUp = true;
	break;
      }
    }
    if (!needCleanUp)
      break;
  } 
  for (it = chowClass.begin(); it!=chowClass.end();it++){
    vector<int> monomial = (*it).first;
    mpz_class coefficient = (*it).second;
    int dimClass = sumVector(monomial);
    if (dimClass > 2*dimSpace || coefficient == 0 || monomial[first_hyperplane] > dimSpace || monomial[second_hyperplane] > dimSpace)
      chowClass.erase(it);
  }

}

 



vector<int> reducedForm(vector<int> monomial){
  if (monomial[exceptional_divisor]  == 0)
      return monomial;
  monomial[first_hyperplane] = monomial[first_hyperplane] +  monomial[second_hyperplane];
  monomial[second_hyperplane] = 0;
  return monomial;
} 


mymap sumOfChowClasses(mymap class1, mymap class2, int dimSpace){
  mymap::iterator it;
  for (it= class1.begin(); it!= class1.end(); it++) {
    vector<int> monomial = (*it).first;
    mpz_class coefficient = (*it).second;
    if (class2.find(monomial) == class2.end())
      class2[monomial] == 0;
    class2[monomial] += coefficient;
  }
  cleanUp(class2,dimSpace);
  return class2;

}

mymap intersectionProduct(mymap class1,mymap class2,  int dimSpace){
  
  mymap result;
  mymap::iterator it1,it2;
  for (it1 =  class1.begin(); it1 != class1.end() ; it1++)
    for  (it2 = class2.begin(); it2 != class2.end(); it2++){
      vector<int> mono1 = (*it1).first;
      vector<int> mono2 = (*it2).first;
      vector<int> product(3,0);
      for (unsigned int i=0;i<product.size();i++)
	product[i] = mono1[i]+mono2[i];
      product = reducedForm(product);
      if  (result.find(product) ==result.end())
	result[product] = 0;
	result[product]+= (*it1).second * (*it2).second;
    }


  cleanUp(result,dimSpace);
  return result;

}




mpz_class topMonomial(int dimE, int dimSpace){
  if (dimE < dimSpace)
    return 0;
  if (dimE == dimSpace)
    return sign(dimSpace-1);
  mpz_class result = 0;
  for(int j=1; j<= dimE-dimSpace;j++)  
    result += topMonomial(dimE-j,dimSpace)*combinations(j,dimSpace+1)*sign(j-1);
  return result;
}


void split(vector<int> v, vector<vector<int> > &subvs) {
	int i;	
	vector<int> curr(v.size(),0);
	while (true) {
		subvs.push_back(curr);
		int id = 0;
		while (id < v.size() && curr[id]== v[id]) 
			id++; 
		if (id ==v.size()) break;
		curr[id]++;
		rep(i,0,id)
			curr[i] = 0;
	}		
}



